
/*********************************************************************************

LED test

measure forward voltage with short current pulses @ nominal current to avoid thermal drift
count and print LED statistics

*********************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <avr/eeprom.h>

#define BINS 200
uint16_t led_bin[BINS];     // LED bins, statistics
int16_t led_cnt;            // LED count
int16_t led_nominal;        // nominal LED voltage /mV
int16_t bin_size;           // bin size / mV
int16_t bin_cnt;            // bin count
int16_t vref;               // ADC reference voltage /mV = VCC with arduino
 
int16_t ee_led_nominal EEMEM;    // nominal LED voltage /mV
int16_t ee_bin_size EEMEM;       // bin size / mV
int16_t ee_bin_cnt EEMEM;        // bin count
int16_t ee_vref EEMEM;           // ADC reference voltage /mV = VCC with arduino

int16_t measure_led(void);
void print_setup(void);
void print_result_line(uint16_t voltage, uint16_t bin);
void print_statistics(void);
int16_t update_stat(uint16_t led_voltage);

void setup() {
  char string[40];

  pinMode(A0, OUTPUT);
  pinMode(A1, OUTPUT);
  digitalWrite(A0, LOW);        // keep pin low, no current through DUT
  digitalWrite(A1, LOW);        // keep pin low, no current through DUT

  Serial.begin(9600);
  Serial.println(F("LED test"));
  Serial.println(F("Press <return> for next measurement"));
  Serial.println(F("Press <space> + <return> for end of measurment and print of statistics"));
  Serial.println(F("Measure the voltage between 5V and GND precisely and enter the value,"));
  Serial.println(F("it is used as the reference for the AD converter on most arduinos."));

  led_nominal = eeprom_read_word((const uint16_t*)&ee_led_nominal);       // nominal LED voltage /mV
  bin_size = eeprom_read_word((const uint16_t*)&ee_bin_size);             // bin size / mV
  bin_cnt = eeprom_read_word((const uint16_t*)&ee_bin_cnt);               // bin count
  vref = eeprom_read_word((const uint16_t*)&ee_vref);                     // ADC reference voltage /mV = VCC with arduino

  print_setup();
  Serial.println(F("Change setup y/n?"));
  while (!Serial.available());

  if (Serial.read()=='y') {
    Serial.setTimeout(30000);
    Serial.print(F("Nominal LED voltage /mV   : "));
    led_nominal = Serial.parseInt();
    if (led_nominal > 5000) led_nominal = 5000;
    Serial.print(F("\nBin size /mV              : "));
    bin_size = Serial.parseInt();
    if (bin_size > 5000) bin_size = 5000;
    Serial.print(F("\nBin count                 : "));
    bin_cnt = Serial.parseInt();
    if (bin_cnt > BINS) bin_cnt = BINS; 
    Serial.print(F("\nADC reference voltage /mV : "));
    vref = Serial.parseInt();
    if (vref >6000) vref = 6000;
    Serial.println(' ');

    eeprom_write_word((uint16_t*)&ee_led_nominal, (uint16_t)led_nominal);       // nominal LED voltage /mV
    eeprom_write_word((uint16_t*)&ee_bin_size, (uint16_t)bin_size);             // bin size / mV
    eeprom_write_word((uint16_t*)&ee_bin_cnt, (uint16_t)bin_cnt);               // bin count
    eeprom_write_word((uint16_t*)&ee_vref, (uint16_t)vref);                     // ADC reference voltage /mV = VCC with arduino
   
    print_setup();
  }
  delay(10);
  while (Serial.available()) Serial.read();    // clear input buffer

  Serial.println(F("#LED ; ULED ; Bin"));
  analogRead(A0);   // dummy read
}

void loop() {
  char cmd;
  int16_t led_voltage, led_bin;

  if (Serial.available()) {  
    cmd = Serial.read();
    switch (cmd) {
      case '\r':    // return, measure next LED
        led_voltage = measure_led();
        led_cnt++;
        led_bin = update_stat(led_voltage); 
        print_result_line(led_voltage, led_bin);
      break;

      case ' ':    // SPACE, finish measurement, print statistics
        print_statistics();
        Serial.println(F("Measurement finished, press reset to restart"));
        while(1); // stop here
      break;

    }
  }
}

int16_t measure_led() {
  int16_t voltage, adc;

  pinMode(A0, INPUT);
  digitalWrite(A1, HIGH);        // power ON for high current version
  delay(1);
  adc = analogRead(A0);
  pinMode(A0, OUTPUT);
  digitalWrite(A0, LOW);
  digitalWrite(A1, LOW);

  voltage = (uint32_t)adc * vref / 1024;

  return voltage;
}

void print_statistics() {
  char string[100];
  int i, j, k, min, max, sum;
  int start=0, end=1;
  long expectancy, sum_voltage;
  int median_voltage=0;

  sprintf_P(string, PSTR("LED statistics\r\n"), led_nominal);
  Serial.print(string);
  sprintf_P(string, PSTR("Bin ;  Min ;  Max ; Count ;   Sum ; Percent graph\r\n"));
  Serial.print(string);
  
  sum = 0;
  sum_voltage=0;
  min = led_nominal - (bin_cnt * bin_size / 2);
  for(i=0; i<bin_cnt; i++, min+=bin_size) { 
    max = min + bin_size - 1;
    sum += led_bin[i];
    sum_voltage += led_bin[i]*(min + (bin_size / 2));
    if (median_voltage==0 && sum >= led_cnt/2) median_voltage=min + (bin_size / 2);
    if (start==0 && sum!=0) start=1;
    if (start==1 && end==1) {
      sprintf_P(string, PSTR("%3u ; %4u ; %4u ; %5u ; %5u ; "), i, min, max, led_bin[i], sum);
      Serial.print(string);
      // pseudo graphical bar, 1% per *
      k = led_bin[i] * 100L / led_cnt;
      for(j=0; j<k; j++) {
        Serial.print('*');
      } 
      Serial.println("");
    }
    if (end==1 && sum == led_cnt) end=0;
  }
  expectancy = sum_voltage / led_cnt;

  sum_voltage = 0;
  min = led_nominal - ((bin_cnt-1) * bin_size / 2);
  for(i=0; i<bin_cnt; i++, min+= bin_size) { 
    sum_voltage += led_bin[i]*(min-expectancy)*(min-expectancy);
  }
  sum_voltage /= led_cnt;

  Serial.println("");
  sprintf_P(string, PSTR("Median             ; %4d ; mV\r\n"), median_voltage);
  Serial.print(string);
  sprintf_P(string, PSTR("Expectancy         ; %4d ; mV\r\n"), expectancy);
  Serial.print(string);
  sprintf_P(string, PSTR("Standard deviation ; %4d ; mV\r\n"), (int)sqrt(sum_voltage));
  Serial.print(string);
}

void print_setup() {
  char string[40];

  sprintf_P(string, PSTR("Nominal LED voltage /mV   : %4u\r\n"), led_nominal);
  Serial.print(string);

  sprintf_P(string, PSTR("Bin size /mV              : %4u\r\n"), bin_size);
  Serial.print(string);

  sprintf_P(string, PSTR("Bin count                 : %4u\r\n"), bin_cnt);
  Serial.print(string);

  sprintf_P(string, PSTR("ADC reference voltage /mV : %4u\r\n"), vref);
  Serial.print(string);
}

void print_result_line(int16_t voltage, int16_t bin) {
  char string[40];

  sprintf_P(string, PSTR("%4d ; %4d ; %3d\r\n"), led_cnt, voltage, bin);
  Serial.print(string);
}

// calculate bin and update statistics

int16_t update_stat(int16_t voltage) {
  int16_t bin;    

  if (voltage >= led_nominal) {
    bin = bin_cnt/2 + ((voltage - led_nominal) / bin_size);
  } else {
    bin = bin_cnt/2 + ((voltage - led_nominal - bin_size + 1) / bin_size); 
  }
  if (bin > (bin_cnt-1) ) bin = (bin_cnt-1);
  if (bin < 0 ) bin = 0;
  led_bin[bin]++;

  return bin;
}
